@@ -1,22 +1,32 @@
-################################################################################################################
-# midi.py       Version 1.0     24-Mar-2025
+#######################################################################################
+# midi.py       Version 1.0     05-Sept-2025
 # Trevor Ritchie, Taj Ballinger, and Bill Manaris
 #
-################################################################################################################
+#######################################################################################
 #
 # [LICENSING GOES HERE]
 #
-################################################################################################################
+#######################################################################################
 #
 # This module includes functionality to connect to MIDI devices from input and output.
 #
+#
+# REVISIONS:
+#
+#
+# TODO:
+#  -
+#
+#######################################################################################
 
-import mido
-from math import log
-from gui import Display, DropDownList, Color
-from timer import Timer2
-import atexit   # for clean exit handling
+import mido                                    # provides MIDI input/output and message handling
+from timer import Timer2                       # for scheduling MIDI events
+from gui import Display, DropDownList, Color   # gui elements for MIDI device selection and display
+import PySide6.QtCore as _QtCore               # Qt core utilities for event handling and integration
+from music import freqToNote                   # converts frequency values to MIDI note numbers
+import atexit                                  # ensures cleanup of MIDI resources on program exit
 
+##### Constants #######################################################################
 # some useful MIDI event constants
 ALL_EVENTS = -1        # a special value so we can register a callback for all events
 NOTE_ON    = 144       # 0x90 - Note On
@@ -49,7 +59,8 @@ PITCHBEND_MIN = 0
 PITCHBEND_MAX = 16383
 PITCHBEND_NORMAL = 8192
 
-# holds active MidiIn and MidiOut objects
+##### Globals #########################################################################
+# holds active MidiIn and MidiOut objects (for cleanup)
 _activeMidiInObjects  = []
 _activeMidiOutObjects = []
 
@@ -57,36 +68,8 @@ _activeMidiOutObjects = []
 # - list of tuples to prevent race conditions
 notesCurrentlyPlaying = []
 
-# MIDI note <-> frequency conversion helper functions
-def freqToNote(frequency):
-   """Converts frequency to the closest MIDI note number with pitch bend value
-      for finer control.  A4 corresponds to the note number 69 (concert pitch
-      is set to 440Hz by default).  The default pitch bend range is 4 half tones.
 
-      Returns (note, pitchBend) where note is a MIDI note number (0-127)
-      and pitchBend is in JythonMusic format (-8192 to 8191).
-   """
-   concertPitch = 440.0   # 440Hz
-   bendRange = 4          # 4 half tones (2 below, 2 above)
-
-   x = log(frequency / concertPitch, 2) * 12 + 69
-   note = round(x)
-   pitchBend = round((x - note) * 8192 / bendRange * 2)
-
-   return int(note), int(pitchBend)
-
-
-def noteToFreq(note):
-   """Converts a MIDI note to the corresponding frequency.  A4 corresponds to the note number 69 (concert pitch
-      is set to 440Hz by default).
-   """
-   concertPitch = 440.0   # 440Hz
-   frequency = concertPitch * 2 ** ( (note - 69) / 12.0 )
-
-   return frequency
-
-
-#################### MidiIn ##############################
+#################### MidiIn ###########################################################
 #
 # MidiIn is used to receive input from a MIDI device.
 #
@@ -105,14 +88,30 @@ def noteToFreq(note):
 #   print("MIDI In - Message Type:", msgType, ", Channel:", msgChannel, ", Data 1:", msgData1, ", Data 2:", msgData2)
 #
 # midiIn.onInput( ALL_EVENTS, processMidiEvent )   # register callback function to handle all input MIDI events
-#
-class MidiIn:
+
+class MidiIn(_QtCore.QObject):
+   """
+   MIDI devices (e.g., a MIDI guitar, keyboard, or control surface) generate MIDI messages when played.
+   If you connect them to your computer (via cable), you can use a MidiIn object to read these messages into your program.
+   When called, it shows a GUI display to select one from the available input MIDI devices.
+   """
+
+   # Define a signal for thread-safe callback execution
+   # NOTE: This signal is required so that midi callbacks which may update the gui
+   # are always executed on the main Qt thread. This is necessary because gui.py
+   # (and Qt in general) requires all gui changes to happen on the thread started by QThread,
+   # which is the main thread running the QApplication event loop.
+
+   # Signal carries: (callback_function, msgType, msgChannel, msgData1, msgData2)
+   callbackSignal = _QtCore.Signal(object, int, int, int, int)
 
    def __init__(self, preferredDevice=""):
       """Initialize a MidiIn object for receiving MIDI input from devices."""
+      super().__init__()  # Initialize QObject
+
       self.preferredDevice = preferredDevice  # remember default choice (if any)
 
-      self.display        = None         # holds selection display for available MIDI devices TODO: do we need this
+      self.display        = None         # holds selection display for available MIDI devices
 
       self.waitingToSetup = True         # used to busy-wait until user has selected a MIDI input device
       self.midiDevice = None             # holds MIDI input device asscociated with this instance
@@ -124,27 +123,28 @@ class MidiIn:
       # holds mido port object (for internal use only)
       self._port = None
 
+      # Connect the signal to our callback execution method
+      # This ensures callbacks execute on the main Qt thread when the signal is emitted
+      self.callbackSignal.connect(self._executeCallback)
+
       # prompt the user to connect a MIDI device to this object
       self.selectMidiInput(self.preferredDevice)
 
-      # register this object in the active list for cleanup
-      _activeMidiInObjects.append(self)
 
-   def close(self):
-      '''Callback used by Receiver interface to close MIDI input device to free any resources it is using'''
-      # close the port if it exists
-      if self._port:
-         self._port.close()
-         self._port = None
+   def __str__(self):
+      return f'MidiIn(preferredDevice = {self.preferredDevice})'
+
+
+   def __repr__(self):
+      return str(self)
 
-      # remove from active list
-      if self in _activeMidiInObjects:
-         _activeMidiInObjects.remove(self)
 
    def selectMidiInput(self, preferredDevice=""):
-      """ Opens preferred MIDI device for input. If this device does not exist,
-          creates display with available input MIDI devices and allows to select one.
       """
+      Opens preferred MIDI device for input. If this device does not exist,
+      creates display with available input MIDI devices and allows to select one.
+      """
+
       self.preferredDevice = preferredDevice    # remember default choice (if any)
 
       # get all available input devices
@@ -161,25 +161,30 @@ class MidiIn:
          self.openInputDevice(self.preferredDevice)
 
       else:  # otherwise, create menu with available devices to select
-         # create selection GUI
-         # TODO: make this a display instead of cli
-         self._displayDeviceSelection()
 
-         # # create selection GUI
-         # self.display = Display("Select MIDI Input", 400, 125) # display info to user
+         # get available input devices
+         items = list(self.inputDevices.keys())
+         items.sort()
 
-         # self.display.drawLabel('Select a MIDI input device from the list', 45, 30)
+         if len(items) > 0:   # if availabale inputs exist
+            # create selection display
+            self.display = Display("Select MIDI Input", 400, 125) # display info to user
+            self.display.drawLabel('Select a MIDI input device from the list', 45, 30)
+
+            # create drop down list
+            deviceDropdown = DropDownList(items, self.openInputDevice)
+            self.display.add(deviceDropdown, 40, 50)
+            self.display.setColor( Color(124, 201, 251) )   # set color to shade of blue (for input)
+
+         else:   # no available inputs
+            print("MidiIn: No available MIDI input devices. Please connect a device.")
 
-         # # create dropdown list of available MIDI input devices
-         # items = self.inputDevices.keys()
-         # items.sort()
-         # deviceDropdown = DropDownList(items, self.openInputDevice)
-         # self.display.add(deviceDropdown, 40, 50)
-         # self.display.setColor( Color(124, 201, 251) )   # set color to shade of blue (for input)
 
    # callback for dropdown list - called when user selects a MIDI device
    def openInputDevice(self, selectedItem):
       """Opens the selected MIDI input device."""
+      global _activeMidiInObjects
+
       try:
          # open selected input device
          print(f'MIDI input device set to "{selectedItem}".')   # let user know
@@ -198,49 +203,37 @@ class MidiIn:
          if self.display:
             self.display.close()   # yes, so close it -- no longer needed
 
+         # register this object in the active list for cleanup
+         _activeMidiInObjects.append(self)
+
       except Exception as e:
          print(f"Error opening MIDI device: {e}")
 
-   def _displayDeviceSelection(self):
-      """Display available MIDI input devices and let user select one."""
-      if not self.inputDevices:
-         print("No MIDI input devices found.")
-         self.waitingToSetup = False
-         return
 
-      print("\nSelect a MIDI input device from the list:")
+   def close(self):
+      '''Callback used by Receiver interface to close MIDI input device to free any resources it is using'''
+      # close the port if it exists
+      if self._port:
+         self._port.close()
+         self._port = None
 
-      # create dropdown list of available MIDI input devices
-      items = list(self.inputDevices.keys())
-      items.sort()
+      # remove from active list
+      if self in _activeMidiInObjects:
+         _activeMidiInObjects.remove(self)
 
-      # display options
-      for i, device in enumerate(items):
-         print(f"{i+1}. {device}")
-
-      # get user selection
-      while self.waitingToSetup:
-         try:
-            choice = input("\nSelect a MIDI input device (number): ")
-            idx = int(choice) - 1
-            if 0 <= idx < len(items):
-               self.openInputDevice(items[idx])
-               break
-            else:
-               print("Invalid selection. Please try again.")
-         except ValueError:
-            print("Please enter a number.")
-         except Exception as e:
-            print(f"Error: {e}")
 
    def onNoteOn(self, function):
       """
       Set up a callback function to handle only noteOn MIDI input events.
       """
+      # print(f"DEBUG: Registering onNoteOn callback: {function}")
       # register the callback for NOTE_ON events
       if NOTE_ON not in self.eventHandlers:
          self.eventHandlers[NOTE_ON] = []
+
       self.eventHandlers[NOTE_ON].append(function)
+      # print(f"DEBUG: onNoteOn callback registered. Total handlers for NOTE_ON: {len(self.eventHandlers[NOTE_ON])}")
+
 
    def onNoteOff(self, function):
       """
@@ -249,8 +242,10 @@ class MidiIn:
       # register the callback for NOTE_OFF events
       if NOTE_OFF not in self.eventHandlers:
          self.eventHandlers[NOTE_OFF] = []
+
       self.eventHandlers[NOTE_OFF].append(function)
 
+
    def onSetInstrument(self, function):
       """
       Set up a callback function to handle only setInstrument MIDI input events.
@@ -260,6 +255,7 @@ class MidiIn:
          self.eventHandlers[SET_INSTRUMENT] = []
       self.eventHandlers[SET_INSTRUMENT].append(function)
 
+
    def onInput(self, eventType, function):
       """
       Associates an incoming event type with a callback function.  Can be used repeatedly to associate different
@@ -271,8 +267,10 @@ class MidiIn:
       # register the callback for the specified event type
       if eventType not in self.eventHandlers:
          self.eventHandlers[eventType] = []
+
       self.eventHandlers[eventType].append(function)
 
+
    def showMessages(self):
       """
       Turns on printing of incoming MIDI messages (useful for exploring what MIDI messages
@@ -280,18 +278,21 @@ class MidiIn:
       """
       self.showIncomingMessages = True
 
+
    def hideMessages(self):
       """
       Turns off printing of incoming MIDI messages.
       """
       self.showIncomingMessages = False
 
+
    def send(self, message, timeStamp):
       """Callback used by Receiver interface to handle incoming MIDI messages."""
       # TODO: do we need this for compatibility?
       # in our implementation, messages are handled by _handleMidiMessage
       pass
 
+
    def _handleMidiMessage(self, message):
       """Handle incoming MIDI messages and dispatch to registered callbacks."""
       # default values
@@ -378,11 +379,13 @@ class MidiIn:
          # some system messages might have data parameters, others don't
          if hasattr(message, 'data1'):
             msgData1 = message.data1
+
          else:
             msgData1 = 0
 
          if hasattr(message, 'data2'):
             msgData2 = message.data2
+
          else:
             msgData2 = 0
 
@@ -406,24 +409,66 @@ class MidiIn:
       if self.showIncomingMessages:
          print(f"{self.midiDeviceName} (MidiIn) - Event Type: {msgType}, Channel: {msgChannel}, Data 1: {msgData1}, Data 2: {msgData2}")
 
+      # print(f"DEBUG: Processing MIDI message - msgType: {msgType}, handlers available: {list(self.eventHandlers.keys())}")
+
       # call event-specific handlers
       if msgType in self.eventHandlers:
-         for handler in self.eventHandlers[msgType]:
+         # print(f"DEBUG: Found {len(self.eventHandlers[msgType])} handlers for msgType {msgType}")
+         for i, handler in enumerate(self.eventHandlers[msgType]):
+            # print(f"DEBUG: Queuing handler {i+1} for msgType {msgType}")
+
             try:
-               handler(msgType, msgChannel, msgData1, msgData2)
+               # THREAD-SAFE CALLBACK QUEUING:
+               # Instead of calling handler() directly (which would run on MIDI thread),
+               # emit a signal that Qt will queue to the main thread's event loop.
+               # This prevents "QBasicTimer can only be used with threads started with QThread" errors
+               # when the callback tries to modify GUI elements.
+               self.callbackSignal.emit(handler, msgType, msgChannel, msgData1, msgData2)
+               # print(f"DEBUG: Successfully queued handler {i+1}")
+
             except Exception as e:
-               print(f"Error in event handler: {e}")
+               print(f"Error queuing MIDI event handler: {e}")
 
       # call general handlers for ALL_EVENTS
       if ALL_EVENTS in self.eventHandlers:
-         for handler in self.eventHandlers[ALL_EVENTS]:
+         # print(f"DEBUG: Found {len(self.eventHandlers[ALL_EVENTS])} handlers for ALL_EVENTS")
+         for i, handler in enumerate(self.eventHandlers[ALL_EVENTS]):
+            # print(f"DEBUG: Queuing ALL_EVENTS handler {i+1}")
+
             try:
-               handler(msgType, msgChannel, msgData1, msgData2)
+               # THREAD-SAFE CALLBACK QUEUING:
+               # Same threading solution as above - emit signal to queue callback on main thread
+               self.callbackSignal.emit(handler, msgType, msgChannel, msgData1, msgData2)
+               # print(f"DEBUG: Successfully queued ALL_EVENTS handler {i+1}")
+
             except Exception as e:
-               print(f"Error in ALL_EVENTS handler: {e}")
+               print(f"Error queuing ALL_EVENTS MIDI handler: {e}")
+
+      # if msgType not in self.eventHandlers and ALL_EVENTS not in self.eventHandlers:
+      #    print(f"DEBUG: No handlers found for msgType {msgType} or ALL_EVENTS")
 
 
-#################### MidiOut ##############################
+   def _executeCallback(self, handler, msgType, msgChannel, msgData1, msgData2):
+      """
+      Execute a MIDI callback with debug information.
+
+      THREADING: This method runs on the main Qt thread (called via signal-slot mechanism).
+      This is safe for GUI operations like adding circles to displays, playing sounds, etc.
+      """
+      # print(f"DEBUG: Executing callback {handler} with args: {msgType}, {msgChannel}, {msgData1}, {msgData2}")
+      try:
+         # Now we can safely call the user's callback function on the main Qt thread
+         # This allows GUI operations like d.add(circle) and Play.noteOn() to work properly
+         handler(msgType, msgChannel, msgData1, msgData2)
+         # print(f"DEBUG: Callback executed successfully")
+
+      except Exception as e:
+         print(f"MidiIn: Error executing callback: {e}")
+         # import traceback
+         # traceback.print_exc()
+
+
+##### MidiOut #########################################################################
 #
 # MidiOut is used to send output a MIDI device.
 #
@@ -450,12 +495,12 @@ class MidiIn:
 # midiOut.sendMidiMessage(noteOff, channel, data1, data2) # end note
 # midiOut.sendMidiMessage(noteOn, channel, data1, 0)      # another way to end note (noteOn with 0 velocity)
 #
+
 class MidiOut:
-   # initialize pitchbend across channels to 0
-   global CURRENT_PITCHBEND
-   CURRENT_PITCHBEND = {}    # holds pitchbend to be used when playing a note / frequency (see below)
-   for i in range(16):
-      CURRENT_PITCHBEND[i] = 0   # set this channel's pitchbend to zero
+   """
+   MIDI devices (e.g., a MIDI synthesizer) can generate sounds from MIDI messages sent to them.
+   If connected to your computer (via cable), you may use a MidiOut object to send such messages from inside your program.
+   """
 
    def __init__(self, preferredDevice=""):
       """Initialize a MidiOut object for sending MIDI output to devices."""
@@ -475,6 +520,11 @@ class MidiOut:
       self.volume = {}            # holds current global volume
       self.panning = {}           # holds current global panning
 
+      # initialize pitchbend across channels to 0
+      self.pitchBend = {}    # holds pitchbend to be used when playing a note / frequency (see below)
+      for i in range(16):
+         self.pitchBend[i] = 0   # set this channel's pitchbend to zero
+
       # initialize all channels
       for channel in range(16):
          self.instrument[channel] = 0   # PIANO (default)
@@ -484,8 +534,16 @@ class MidiOut:
       # prompt the user to connect to a MIDI device
       self.selectMidiOutput(self.preferredDevice)
 
-      # register this object in the active list for cleanup
-      _activeMidiOutObjects.append(self)
+      # # and, since above GUI is asynchronous, wait until user has selected a MIDI output device
+      # while(self.waitingToSetup):
+      #    sleep(0.1)    # sleep for 0.1 second
+
+
+   def __str__(self):
+      return f'MidiOut(preferredDevice = {self.preferredDevice})'
+
+   def __repr__(self):
+      return str(self)
 
    def close(self):
       '''Close MIDI output device to free any resources it is using.'''
@@ -501,13 +559,16 @@ class MidiOut:
       if self in _activeMidiOutObjects:
          _activeMidiOutObjects.remove(self)
 
+
    def play(self, material):
       """Play jMusic material (Score, Part, Phrase, Note) using the MIDI output device."""
       pass
 
+
    def noteOn(self, pitch, velocity=100, channel=0, panning = -1):
       """Send a NOTE_ON message for this pitch to the selected output MIDI device.  Default panning of -1 means to
          use the default (global) panning setting of the selected MIDI device."""
+
       if (type(pitch) == int) and (0 <= pitch <= 127):   # a MIDI pitch?
          # for standard MIDI notes, use the current pitch bend setting
          # if panning is specified, set it
@@ -545,6 +606,7 @@ class MidiOut:
       else:
          print(f"MidiOut.frequencyOn(): Invalid frequency {frequency}, expected frequency in Hz from 8.17 to 12600.0 (float).")
 
+
    def noteOff(self, pitch, channel=0):
       """Send a NOTE_OFF message for this pitch to the selected output MIDI device."""
 
@@ -565,6 +627,7 @@ class MidiOut:
       else:
          print(f"MidiOut.noteOff(): Unrecognized pitch {pitch}, expected MIDI pitch from 0 to 127 (int), or frequency in Hz from 8.17 to 12600.0 (float).")
 
+
    def frequencyOff(self, frequency, channel=0):
       """Send a NOTE_OFF message for this frequency (in Hz) to the selected output MIDI device."""
       if (type(frequency) == float) and (8.17 <= frequency <= 12600.0): # a frequency in Hertz (within MIDI pitch range 0 to 127)?
@@ -585,6 +648,7 @@ class MidiOut:
          # frequency was outside expected range
          print(f"MidiOut.frequencyOff(): Invalid frequency {frequency}, expected frequency in Hz from 8.17 to 12600.0 (float).")
 
+
    def note(self, pitch, start, duration, velocity=100, channel=0, panning = -1):
       """Plays a note with given 'start' time (in milliseconds from now), 'duration' (in milliseconds
          from 'start' time), with given 'velocity' on 'channel'.  Default panning of -1 means to
@@ -605,6 +669,7 @@ class MidiOut:
       #        When the timers elapse, then the two objects (in theory) should be garbage-collectable,
       #        and should be eventually cleaned up.  So, here, no effort is made in reusing timer objects, etc.
 
+
    def frequency(self, frequency, start, duration, velocity=100, channel=0, panning = -1):
       """Plays a frequency with given 'start' time (in milliseconds from now), 'duration' (in milliseconds
          from 'start' time), with given 'velocity' on 'channel'.  Default panning of -1 means to
@@ -626,13 +691,13 @@ class MidiOut:
       frequencyOn.start()
       frequencyOff.start()
 
+
    def setPitchBend(self, bend = 0, channel=0):
       """Set global pitchbend variable to be used when a note / frequency is played.
          Per JythonMusic API: Pitch bend ranges from -8192 (max downward bend) to 8191 (max upward bend).
          No pitch bend is 0 (this is the default).
       """
-      global CURRENT_PITCHBEND
-      CURRENT_PITCHBEND[channel] = bend
+      self.pitchBend[channel] = bend
 
       # Convert from JythonMusic format (-8192 to 8191) to MIDI format (0-16383)
       # where 8192 is the center (no bend)
@@ -649,10 +714,11 @@ class MidiOut:
       # send the MIDI message
       self.sendMidiMessage(PITCH_BEND, channel, lsb, msb)
 
+
    def getPitchBend(self, channel=0):
       """Returns the current pitchbend for this channel."""
-      global CURRENT_PITCHBEND
-      return CURRENT_PITCHBEND[channel]
+      return self.pitchBend[channel]
+
 
    def noteOnPitchBend(self, pitch, bend = 0, velocity=100, channel=0, panning = -1):
       """Send a NOTE_ON message for this pitch and pitch bend to the selected output MIDI device.
@@ -667,10 +733,12 @@ class MidiOut:
 
       self.sendMidiMessage(NOTE_ON, channel, pitch, velocity)  # send the message
 
+
    def allNotesOff(self):
       """Turns off all notes on all channels."""
       self.allFrequenciesOff()
 
+
    def allFrequenciesOff(self):
       """Turns off all frequencies on all channels."""
       for channel in range(16):  # cycle through all channels
@@ -680,6 +748,7 @@ class MidiOut:
          # also reset pitch bend
          self.setPitchBend(0, channel)
 
+
    def stop(self):
       """Stops all MIDI music from sounding."""
       # NOTE:  This could also handle self.note() notes, which may have been
@@ -696,6 +765,7 @@ class MidiOut:
       # NOTE: In the future, we may also want to handle scheduled notes through self.note().  This could be done
       # by creating a list of Timers created via note() and looping through them to stop them here.
 
+
    def setInstrument(self, instrument, channel=0):
       """Sets 'channel' to 'instrument' for this channel of the output MIDI device."""
 
@@ -703,30 +773,36 @@ class MidiOut:
       self.sendMidiMessage(SET_INSTRUMENT, channel, instrument, 0)   # and set it
       # (see controller numbers - http://www.indiana.edu/~emusic/cntrlnumb.html)
 
+
    def getInstrument(self, channel=0):
       """Gets the current instrument for this channel of the output MIDI device."""
       return self.instrument[channel]
 
+
    def setVolume(self, volume, channel=0):
       """Sets the current coarse volume for this channel of the output MIDI device."""
       self.volume[channel] = volume                               # remember it
       self.sendMidiMessage(CONTROL_CHANGE, channel, 7, volume)    # and set it (7 is controller number for global / main volume)
       # (see controller numbers - http://www.indiana.edu/~emusic/cntrlnumb.html)
 
+
    def getVolume(self, channel=0):
       """Gets the current coarse volume for this channel of the output MIDI device."""
       return self.volume[channel]
 
+
    def setPanning(self, panning, channel=0):
       """Sets the current panning setting for this channel of the output MIDI device."""
       self.panning[channel] = panning                              # remember it
       self.sendMidiMessage(CONTROL_CHANGE, channel, 10, panning)   # then, use it (otherwise let default / global panning stand)
       # (see controller numbers - http://www.indiana.edu/~emusic/cntrlnumb.html)
 
+
    def getPanning(self, channel=0):
       """Gets the current panning setting for this channel of the output MIDI device."""
       return self.panning[channel]
 
+
    ####### function to output MIDI message through selected output MIDI device ########
    def sendMidiMessage(self, msgType, msgChannel, msgData1, msgData2):
       """Send a MIDI message with the given parameters."""
@@ -828,57 +904,30 @@ class MidiOut:
          self.openOutputDevice(self.preferredDevice)
 
       else:  # otherwise, create menu with available devices to select
-         # create selection GUI
-         # TODO: make this a display instead of cli
-         self._displayDeviceSelection()
 
-         # # create selection GUI
-         # self.display = Display("Select MIDI Output", 400, 125) # display info to user
+         # get available output devices
+         items = list(self.outputDevices.keys())
+         items.sort()
 
-         # self.display.drawLabel('Select a MIDI output device from the list', 45, 30)
+         if len(items) > 0:   # if available outputs exist
+            # create selection display
+            self.display = Display("Select MIDI Output", 400, 125) # display info to user
+            self.display.drawLabel('Select a MIDI output device from the list', 45, 30)
 
-         # # create dropdown list of available MIDI input devices
-         # items = self.outputDevices.keys()
-         # items.sort()
-         # deviceDropdown = DropDownList(items, self.openOutputDevice)
-         # self.display.add(deviceDropdown, 40, 50)
-         # self.display.setColor( Color(255, 153, 153) )   # set color to shade of red (for output)
+            # create drop down list
+            deviceDropdown = DropDownList(items, self.openOutputDevice)
+            self.display.add(deviceDropdown, 40, 50)
+            self.display.setColor( Color(255, 153, 153) )   # set color to shade of red (for output)
 
-   def _displayDeviceSelection(self):
-      """Display available MIDI output devices and let user select one."""
-      if not self.outputDevices:
-         print("No MIDI output devices found.")
-         self.waitingToSetup = False
-         return
+         else:   # no available outputs
+            print("MidiOut: No available MIDI output devices.")
 
-      print("\nSelect a MIDI output device from the list:")
-
-      # create dropdown list of available MIDI output devices
-      items = list(self.outputDevices.keys())
-      items.sort()
-
-      # display options
-      for i, device in enumerate(items):
-         print(f"{i+1}. {device}")
-
-      # get user selection
-      while self.waitingToSetup:
-         try:
-            choice = input("\nSelect a MIDI output device (number): ")
-            idx = int(choice) - 1
-            if 0 <= idx < len(items):
-               self.openOutputDevice(items[idx])
-               break
-            else:
-               print("Invalid selection. Please try again.")
-         except ValueError:
-            print("Please enter a number.")
-         except Exception as e:
-            print(f"Error: {e}")
 
    # callback for dropdown list
    def openOutputDevice(self, selectedItem):
       """Opens the selected MIDI output device."""
+      global _activeMidiOutObjects
+
       try:
          # open selected output device
          print(f'MIDI output device set to "{selectedItem}".')   # let user know
@@ -897,9 +946,13 @@ class MidiOut:
          if self.display:
             self.display.close()   # yes, so close it -- no longer needed
 
+         # register this object in the active list for cleanup
+         _activeMidiOutObjects.append(self)
+
       except Exception as e:
          print(f"Error opening MIDI device: {e}")
 
+
 # function to stop and clean-up all active Midi objects
 def _stopActiveMidiObjects():
    """Clean up all active MIDI objects when program is stopped."""
@@ -912,7 +965,7 @@ def _stopActiveMidiObjects():
    # close all MidiIn objects
    for midiIn in midiInObjects:
       try:
-         print(f"Closing MIDI input device: {midiIn.midiDeviceName}")
+         # print(f"Closing MIDI input device: {midiIn.midiDeviceName}")
          midiIn.close()
       except Exception as e:
          print(f"Error closing MIDI input device: {e}")
@@ -920,7 +973,7 @@ def _stopActiveMidiObjects():
    # close all MidiOut objects
    for midiOut in midiOutObjects:
       try:
-         print(f"Closing MIDI output device: {midiOut.midiDeviceName}")
+         # print(f"Closing MIDI output device: {midiOut.midiDeviceName}")
          # first stop all notes playing
          midiOut.allNotesOff()
          midiOut.close()
@@ -931,12 +984,13 @@ def _stopActiveMidiObjects():
    _activeMidiInObjects.clear()
    _activeMidiOutObjects.clear()
 
-   print("All MIDI devices closed.")
+   # print("All MIDI devices closed.")
 
 # register the cleanup function to be called when the program exits
 atexit.register(_stopActiveMidiObjects)
 
-#################### Unit Testing ##############################
+
+##### Tests ###########################################################################
 
 if __name__ == '__main__':
    pass
